import os
import pandas as pd
from sklearn.preprocessing import StandardScaler
import logging
from sklearn.model_selection import train_test_split
from feature_engineering import add_features
import sys

# 프로젝트 루트 및 스크립트 경로 설정
project_root = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..'))
scripts_dir = os.path.join(project_root, 'scripts')
config_dir = os.path.join(project_root, 'config')
utils_dir = os.path.join(scripts_dir, 'utils')

sys.path.extend([
    scripts_dir,
    os.path.join(scripts_dir, 'data_collection'),
    os.path.join(scripts_dir, 'data_processing'),
    os.path.join(scripts_dir, 'model'),
    os.path.join(scripts_dir, 'trading'),
    utils_dir
])

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'scripts', 'model')))

logging.basicConfig(level=logging.INFO)

def normalize_data(df):
    """Normalize the data using StandardScaler."""
    scaler = StandardScaler()
    normalized_df = pd.DataFrame(scaler.fit_transform(df), columns=df.columns)
    return normalized_df, scaler

def load_csv_files_from_directory(directory_path, identifier_col=None):
    dataframes = []
    for file_name in os.listdir(directory_path):
        if file_name.endswith('.csv'):
            file_path = os.path.join(directory_path, file_name)
            try:
                df = pd.read_csv(file_path, encoding='utf-8')
                logging.info(f"Loaded {file_name} with columns: {df.columns}")
                df.columns = df.columns.str.lower()  # Make all column names lowercase

                if identifier_col and identifier_col not in df.columns:
                    # 파일명에서 식별자를 추출 (예: 파일명에서 종목코드를 추출하여 컬럼으로 추가)
                    ticker = os.path.splitext(file_name)[0].split('_')[0]
                    df[identifier_col] = ticker

                if not df.empty:
                    logging.info(f"First few rows of {file_name}:\n{df.head()}")
                    dataframes.append(df)
                else:
                    logging.warning(f"Empty file: {file_path}")
            except pd.errors.EmptyDataError:
                logging.warning(f"Empty file: {file_path}")
            except Exception as e:
                logging.error(f"Error reading {file_path}: {e}")
    if dataframes:
        return pd.concat(dataframes, ignore_index=True)
    else:
        logging.error(f"No valid CSV files found in directory: {directory_path}")
        return pd.DataFrame()

from tensorflow.keras.models import load_model

def update_lstm_model(model_path, X_train, y_train, epochs=5, batch_size=32):
    """
    기존 LSTM 모델을 불러와 추가 학습을 수행합니다.

    Parameters:
    - model_path (str): 학습된 모델 파일 경로
    - X_train (np.array): 학습 데이터
    - y_train (np.array): 학습 타겟
    - epochs (int): 학습 에포크 수
    - batch_size (int): 배치 크기

    Returns:
    - model: 추가 학습이 완료된 모델
    """
    # 기존 모델 불러오기
    model = load_model(model_path)
    
    # 추가 학습 수행
    model.fit(X_train, y_train, epochs=epochs, batch_size=batch_size)
    
    return model


def preprocess_df(df, date_column='date', date_format=None):
    if date_column not in df.columns:
        raise ValueError(f"Date column '{date_column}' not found in the dataset. Available columns: {df.columns}")
    
    logging.info(f"Before date conversion, first few rows:\n{df.head()}")

    # 날짜 형식이 지정되면, 해당 형식으로 변환
    try:
        if date_format:
            df[date_column] = pd.to_datetime(df[date_column], format=date_format, errors='coerce')
        else:
            df[date_column] = pd.to_datetime(df[date_column], errors='coerce')
    except Exception as e:
        logging.error(f"Date conversion error: {e}")
        return pd.DataFrame()

    logging.info(f"After date conversion, first few rows:\n{df.head()}")

    # 날짜가 변환되지 않은 (NaT로 남은) 행들을 제거
    df.dropna(subset=[date_column], inplace=True)

    logging.info(f"After dropping rows with invalid dates, shape: {df.shape}")

    df.fillna(0, inplace=True)
    return df

def preprocess_data(stock_data, financial_data, economic_data):
    if stock_data.empty:
        logging.error("Stock data is empty. Cannot proceed with preprocessing.")
        return None
    
    logging.info(f"Initial stock data shape: {stock_data.shape}")
    
    # 주식 데이터는 '%Y-%m-%d' 형식으로 변환
    stock_data = preprocess_df(stock_data, date_column='date', date_format='%Y-%m-%d')
    
    if stock_data.empty:
        logging.error("Preprocessed stock data is empty after date conversion.")
        return None
    
    logging.info(f"Stock data after preprocessing shape: {stock_data.shape}")
    
    # 재무 데이터는 '%Y%m%d' 형식으로 변환 (이 부분은 데이터셋에 맞춰 변경)
    if not financial_data.empty:
        financial_data = preprocess_df(financial_data, date_column='date', date_format='%Y%m%d')
        if financial_data.empty:
            logging.warning("Financial data is empty after preprocessing.")
    
    # 경제 데이터는 '%Y%m%d' 형식으로 변환 (필요에 따라 맞춤 설정)
    if not economic_data.empty:
        economic_data = preprocess_df(economic_data, date_column='date', date_format='%Y%m%d')
        if economic_data.empty:
            logging.warning("Economic data is empty after preprocessing.")

    # 목표 값 생성
    stock_data['target'] = stock_data['close'].shift(-1)
    stock_data.dropna(subset=['target'], inplace=True)

    if stock_data.empty:
        logging.error("No valid stock data after creating target.")
        return None

    # 데이터 병합 단계 (종목별로 재무 데이터와 병합)
    if not financial_data.empty:
        stock_data = pd.merge_asof(stock_data.sort_values('date'), financial_data.sort_values('date'), on='date', direction='nearest')
        logging.info(f"Stock data shape after merging with financial data: {stock_data.shape}")
    
    # 경제 데이터를 공통적으로 병합
    if not economic_data.empty:
        stock_data = pd.merge_asof(stock_data.sort_values('date'), economic_data.sort_values('date'), on='date', direction='nearest')
        logging.info(f"Stock data shape after merging with economic data: {stock_data.shape}")

    if stock_data.empty:
        logging.error("No valid combined data after merging.")
        return None

    # 피처 엔지니어링
    stock_data = add_features(stock_data)

    # Object 타입의 컬럼 처리 및 데이터 정제
    for col in stock_data.columns:
        if stock_data[col].dtype == 'object':
            stock_data[col] = pd.to_numeric(stock_data[col].str.replace(',', ''), errors='coerce')

    stock_data.ffill(inplace=True)
    stock_data.bfill(inplace=True)
    stock_data.fillna(0, inplace=True)

    # 피처와 타겟 분리
    X = stock_data.drop(['target', 'date'], axis=1)
    y = stock_data['target']

    if X.empty or y.empty:
        logging.error("Insufficient data after preprocessing for training the model.")
        return None

    # 데이터 분할 및 스케일링
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=42)
    scaler = StandardScaler()
    X_train = scaler.fit_transform(X_train)
    X_test = scaler.transform(X_test)

    return (X_train, X_test, y_train, y_test), scaler, stock_data


def preprocess_and_save(stock_data, financial_data, economic_data, output_dir, corp_code):
    result = preprocess_data(stock_data, financial_data, economic_data)
    
    if result is None:
        logging.error(f"Data preprocessing failed for {corp_code}.")
        return

    (_, _, _, _), _, combined_data = result

    if combined_data is None or combined_data.empty:
        logging.error(f"No valid combined data after preprocessing for {corp_code}.")
        return

    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    # 파일명을 종목 코드가 포함된 형태로 지정하여 저장
    output_path = os.path.join(output_dir, f'preprocessed_data_{corp_code}.csv')
    combined_data.to_csv(output_path, index=False)
    logging.info(f"Preprocessed data for {corp_code} saved to {output_path}")


def main(base_dir):
    stock_data_dir = os.path.join(base_dir, 'stock_data')
    financial_data_dir = os.path.join(base_dir, 'financial_data')
    economic_data_dir = os.path.join(base_dir, 'economic_data')

    # 주식 데이터를 로드할 때 식별 컬럼을 추가
    stock_data = load_csv_files_from_directory(stock_data_dir, identifier_col='corp_code')
    if stock_data.empty:
        logging.error("Stock data is missing or empty after loading.")
        return
    logging.info(f"Stock data initial state: {stock_data.head()}")

    financial_data = load_csv_files_from_directory(financial_data_dir)
    economic_data = load_csv_files_from_directory(economic_data_dir)

    if financial_data.empty:
        logging.warning("Financial data is missing or empty.")
    if economic_data.empty:
        logging.warning("Economic data is missing or empty.")
    
        # Iterate over each corp_code (종목 코드로 묶어 전처리)
    identifier_column = 'corp_code'
    if identifier_column in stock_data.columns:
        corp_codes = stock_data[identifier_column].unique()
    else:
        logging.error(f"'{identifier_column}' column not found. Cannot proceed without proper identification information.")
        return

    output_dir = os.path.join(base_dir, 'preprocessed_data')  # 한 폴더에 저장

    for corp_code in corp_codes:
        corp_stock_data = stock_data[stock_data[identifier_column] == corp_code]
        corp_financial_data = financial_data[financial_data[identifier_column] == corp_code] if not financial_data.empty else pd.DataFrame()

        # Preprocess and save data for each corp_code
        preprocess_and_save(corp_stock_data, corp_financial_data, economic_data, output_dir, corp_code)

if __name__ == "__main__":
    base_dir = 'C:\\Users\\user\\projects\\st\\data'  # Use your data directory path
    main(base_dir)

