import os
import pandas as pd
import numpy as np
from sklearn.preprocessing import StandardScaler
from sklearn.metrics import mean_squared_error, mean_absolute_error, r2_score
import tensorflow as tf
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import LSTM, Dense, Dropout, Bidirectional
from tensorflow.keras.callbacks import EarlyStopping, ModelCheckpoint
from keras_tuner import RandomSearch  # Keras Tuner (패키지에 따라 'keras_tuner' 또는 'kt-keras-tuner')

import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import logging
import joblib
from scipy.ndimage import shift
from autokeras import StructuredDataRegressor

# 로그 설정
logging.basicConfig(level=logging.INFO)

# 데이터 경로 설정
DATA_DIR = "C:/Users/user/projects/st/data"
PREPROCESSED_DATA_DIR = os.path.join(DATA_DIR, "preprocessed_data")
MODEL_DIR = os.path.join(DATA_DIR, "models")
PLOT_DIR = os.path.join(DATA_DIR, "plots")
os.makedirs(PLOT_DIR, exist_ok=True)

SEQUENCE_LENGTH = 60

def load_and_preprocess_data(file_path):
    df = pd.read_csv(file_path)
    df['date'] = pd.to_datetime(df['date'])

    # 거래 정지된 날짜 확인 및 처리
    df['volume'].replace(0, np.nan, inplace=True)
    df['close'].replace(0, np.nan, inplace=True)
    df.fillna(method='ffill', inplace=True)  # 이전 값으로 결측값 채우기

    # 비정상적 데이터 필터링: 최소한의 거래량 조건 추가
    df = df[df['volume'] > 0]

    # 기타 전처리
    scaler = StandardScaler()
    feature_columns = [col for col in df.columns if col not in ['date']]
    df[feature_columns] = scaler.fit_transform(df[feature_columns])
    
    # 데이터 나누기
    train_data = df[:int(0.7 * len(df))]
    val_data = df[int(0.7 * len(df)):int(0.85 * len(df))]
    test_data = df[int(0.85 * len(df)):]

    return train_data, val_data, test_data, feature_columns, scaler

def augment_data(X, y):
    # 데이터 시프트
    X_shifted = shift(X, shift=[0, 1, 0], mode='nearest')
    y_shifted = y

    # 원본 데이터와 결합하여 증강 데이터 생성
    X_augmented = np.concatenate([X, X_shifted], axis=0)
    y_augmented = np.concatenate([y, y_shifted], axis=0)

    return X_augmented, y_augmented

def create_sequences(data, seq_length, feature_columns):
    X, y = [], []
    for i in range(len(data) - seq_length):
        X.append(data.iloc[i:i+seq_length][feature_columns].values)
        y.append(data.iloc[i+seq_length]['target'])
    return np.array(X), np.array(y)

def build_model(hp, feature_columns, SEQUENCE_LENGTH):
    model = Sequential()
    model.add(Bidirectional(LSTM(units=hp.Int('units_lstm1', min_value=32, max_value=128, step=32), return_sequences=True),
                            input_shape=(SEQUENCE_LENGTH, len(feature_columns))))
    model.add(Dropout(rate=hp.Float('dropout_1', min_value=0.1, max_value=0.5, step=0.1)))
    
    model.add(LSTM(units=hp.Int('units_lstm2', min_value=32, max_value=128, step=32), return_sequences=False))
    model.add(Dropout(rate=hp.Float('dropout_2', min_value=0.1, max_value=0.5, step=0.1)))
    
    model.add(Dense(1, activation='linear'))
    
    model.compile(optimizer=tf.keras.optimizers.Adam(hp.Choice('learning_rate', values=[1e-2, 1e-3, 1e-4])),
                  loss='mse')
    
    return model


def tune_and_train_model(X_train, y_train, X_val, y_val, feature_columns, SEQUENCE_LENGTH):
    tuner = RandomSearch(
        lambda hp: build_model(hp, feature_columns, SEQUENCE_LENGTH),  # feature_columns와 SEQUENCE_LENGTH 전달
        objective='val_loss',
        max_trials=5,
        executions_per_trial=1,
        directory='tuner_results',
        project_name='stock_prediction'
    )

    early_stop = EarlyStopping(monitor='val_loss', patience=10, restore_best_weights=True)
    
    tuner.search(X_train, y_train, epochs=50, validation_data=(X_val, y_val), callbacks=[early_stop])
    
    best_model = tuner.get_best_models(num_models=1)[0]
    
    return best_model, tuner


def evaluate_model(model, X_test, y_test, corp_code):
    y_pred = model.predict(X_test)

    mse = mean_squared_error(y_test, y_pred)
    mae = mean_absolute_error(y_test, y_pred)
    r2 = r2_score(y_test, y_pred)

    print(f"Test MSE: {mse:.4f}")
    print(f"Test MAE: {mae:.4f}")
    print(f"Test R2 Score: {r2:.4f}")

    plt.figure(figsize=(12,6))
    plt.plot(y_test, label='Actual')
    plt.plot(y_pred, label='Predicted')
    plt.title('Actual vs Predicted Target Values')
    plt.xlabel('Time')
    plt.ylabel('Target')
    plt.legend()

    # 플롯 범위 설정에 대한 예외 처리
    try:
        ymin = min(y_test.min(), y_pred.min())
        ymax = max(y_test.max(), y_pred.max())
        if not np.isfinite(ymin) or not np.isfinite(ymax) or ymin == ymax:
            ymin, ymax = 0, 1  # 기본값 설정
        plt.ylim(ymin, ymax)
        plt.xlim(0, len(y_test))
    except ValueError as e:
        print(f"Error in setting plot limits: {e}")
        plt.ylim(0, 1)  # 기본값으로 설정

    ax = plt.gca()
    ax.xaxis.set_major_locator(ticker.MaxNLocator(integer=True))
    ax.yaxis.set_major_locator(ticker.MaxNLocator(integer=True))

    plot_path = os.path.join(PLOT_DIR, f'{corp_code}_actual_vs_predicted.png')
    plt.savefig(plot_path)
    plt.close()

    return y_pred

def calculate_and_compare_returns(test_data, y_pred, corp_code):
    try:
        actual_prices = test_data.iloc[SEQUENCE_LENGTH:]['close'].values
        signal = np.where(y_pred[:-1] < y_pred[1:], 1, -1)  # 매수/매도 신호 생성
        returns = np.diff(actual_prices) / actual_prices[:-1]
        strategy_returns = returns * signal
        cumulative_returns = np.cumprod(1 + strategy_returns) - 1
        benchmark_returns = np.cumprod(1 + returns) - 1

        plt.figure(figsize=(12, 6))
        plt.plot(cumulative_returns, label='Strategy Returns')
        plt.plot(benchmark_returns, label='Benchmark Returns')
        plt.title('Cumulative Returns Comparison')
        plt.xlabel('Time')
        plt.ylabel('Cumulative Returns')
        plt.legend()

        # 축이 유효한 범위 내에 있는지 확인
        plt.ylim(bottom=min(cumulative_returns.min(), benchmark_returns.min()), top=max(cumulative_returns.max(), benchmark_returns.max()))

        plot_path = os.path.join(PLOT_DIR, f'{corp_code}_cumulative_returns.png')
        plt.savefig(plot_path)
        plt.close()  # Close the plot to avoid displaying it

        print(f"Strategy Final Return: {cumulative_returns[-1] * 100:.2f}%")
        print(f"Benchmark Final Return: {benchmark_returns[-1] * 100:.2f}%")
    except Exception as e:
        print(f"An error occurred while calculating returns for {corp_code}: {e}")
        # 오류가 발생한 경우 기본 결과값을 사용할 수 있도록 추가 처리
        cumulative_returns, benchmark_returns = [0], [0]

    return cumulative_returns, benchmark_returns


def run_automl(X_train, y_train, X_val, y_val):
    # AutoML을 사용하여 모델 최적화
    automl_model = StructuredDataRegressor(max_trials=10)
    automl_model.fit(X_train, y_train, validation_data=(X_val, y_val), epochs=50)
    return automl_model

def main():
    for file_name in os.listdir(PREPROCESSED_DATA_DIR):
        if file_name.endswith('.csv'):
            corp_code = os.path.splitext(file_name)[0]
            file_path = os.path.join(PREPROCESSED_DATA_DIR, file_name)

            logging.info(f"Processing {corp_code}...")

            train_data, val_data, test_data, feature_columns, scaler = load_and_preprocess_data(file_path)
            X_train, y_train = create_sequences(train_data, SEQUENCE_LENGTH, feature_columns)
            X_val, y_val = create_sequences(val_data, SEQUENCE_LENGTH, feature_columns)
            X_test, y_test = create_sequences(test_data, SEQUENCE_LENGTH, feature_columns)

            # 데이터 증강 적용
            X_train, y_train = augment_data(X_train, y_train)

            # 모델 성능 평가 및 루프 개선
            model, tuner = tune_and_train_model(X_train, y_train, X_val, y_val, feature_columns, SEQUENCE_LENGTH)
            y_pred = evaluate_model(model, X_test, y_test, corp_code)
            
            if r2_score(y_test, y_pred) < 0.8:  # R2 스코어가 0.8 미만일 경우 AutoML 실행
                model = run_automl(X_train, y_train, X_val, y_val)

            calculate_and_compare_returns(test_data, y_pred, corp_code)

            corp_model_dir = os.path.join(MODEL_DIR, corp_code)
            os.makedirs(corp_model_dir, exist_ok=True)
            model_path = os.path.join(corp_model_dir, f"{corp_code}_lstm_model.h5")
            scaler_path = os.path.join(corp_model_dir, f"{corp_code}_scaler.pkl")

            model.save(model_path)
            joblib.dump(scaler, scaler_path)
            logging.info(f"Model and scaler saved for {corp_code}.")


if __name__ == "__main__":
    main()
